# Copyright (C) 2019-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Script to update the year of last modification
in the license header of source code files.\n
Also it appends a header if it is missing in the file.
"""

import re
import sys
from argparse import ArgumentParser, FileType, Namespace
from datetime import datetime
from pathlib import Path
from subprocess import CalledProcessError, run
from typing import Dict, List, Tuple, Union

from pontos.terminal import Terminal
from pontos.terminal.null import NullTerminal
from pontos.terminal.rich import RichTerminal

SUPPORTED_FILE_TYPES = [
    ".bash",
    ".c",
    ".h",
    ".go",
    ".cmake",
    ".js",
    ".nasl",
    ".po",
    ".py",
    ".sh",
    ".txt",
    ".xml",
    ".xsl",
]
SUPPORTED_LICENCES = [
    "AGPL-3.0-or-later",
    "GPL-2.0-only",
    "GPL-2.0-or-later",
    "GPL-3.0-or-later",
]


def _get_modified_year(f: Path) -> str:
    """In case of the changed arg, update year to last modified year"""
    try:
        cmd = ["git", "log", "-1", "--format=%ad", "--date=format:%Y", str(f)]
        proc = run(
            cmd,
            text=True,
            capture_output=True,
            check=True,
            universal_newlines=True,
        )
        return proc.stdout.rstrip()
    except CalledProcessError as e:
        raise e


def _find_copyright(
    line: str,
    regex: re.Pattern,
) -> Tuple[bool, Union[Dict[str, Union[str, None]], None]]:
    """Match the line for the regex"""
    copyright_match = re.search(regex, line)
    if copyright_match:
        return (
            True,
            {
                "creation_year": copyright_match.group(1),
                "modification_year": copyright_match.group(2),
                "company": copyright_match.group(3),
            },
        )
    return False, None


def _add_header(
    suffix: str, licence: str, company: str, year: str
) -> Union[str, None]:
    """Tries to add the header to the file.
    Requirements:
      - file type must be supported
      - licence file must exist
    """
    if suffix in SUPPORTED_FILE_TYPES:
        root = Path(__file__).parent
        licence_file = root / "templates" / licence / f"template{suffix}"
        try:
            return (
                licence_file.read_text(encoding="utf-8")
                .replace("<company>", company)
                .replace("<year>", year)
            )
        except FileNotFoundError as e:
            raise e
    else:
        raise ValueError


def _update_file(
    file: Path,
    regex: re.Pattern,
    parsed_args: Namespace,
    term: Terminal,
) -> None:
    """Function to update the given file.
    Checks if header exists. If not it adds an
    header to that file, else it checks if year
    is up to date
    """

    if parsed_args.changed:
        try:
            parsed_args.year = _get_modified_year(file)
        except CalledProcessError:
            term.warning(
                f"{file}: Could not get date of last modification"
                f" using git, using {str(parsed_args.year)} instead."
            )

    try:
        with file.open("r+") as fp:
            found = False
            i = 10  # assume that copyright is in the first 10 lines
            while not found and i > 0:
                line = fp.readline()
                if line == "":
                    i = 0
                    continue
                found, copyright_match = _find_copyright(line=line, regex=regex)
                i = i - 1
            # header not found, add header
            if i == 0 and not found:
                try:
                    header = _add_header(
                        file.suffix,
                        parsed_args.licence,
                        parsed_args.company,
                        parsed_args.year,
                    )
                    if header:
                        fp.seek(0)  # back to beginning of file
                        rest_of_file = fp.read()
                        fp.seek(0)
                        fp.write(header)
                        fp.write("\n")
                        fp.write(rest_of_file)
                        print(f"{file}: Added licence header.")
                        return 0
                except ValueError:
                    print(
                        f"{file}: No licence header for the"
                        f" format {file.suffix} found.",
                    )
                except FileNotFoundError:
                    print(
                        f"{file}: Licence file for {parsed_args.licence} "
                        "is not existing."
                    )
                return 1
            # replace found header and write it to file
            if (
                not copyright_match["modification_year"]
                and copyright_match["creation_year"] < parsed_args.year
                or copyright_match["modification_year"]
                and copyright_match["modification_year"] < parsed_args.year
            ):
                copyright_term = (
                    f'Copyright (C) {copyright_match["creation_year"]}'
                    f'-{parsed_args.year} {copyright_match["company"]}'
                )
                new_line = re.sub(regex, copyright_term, line)
                fp_write = fp.tell() - len(line)  # save position to insert
                rest_of_file = fp.read()
                fp.seek(fp_write)
                fp.write(new_line)
                fp.write(rest_of_file)
                # in some cases we replace "YYYY - YYYY" with "YYYY-YYYY"
                # resulting in 2 characters left at the end of the file
                # so we truncate the file, just in case!
                fp.truncate()
                print(
                    f"{file}: Changed Licence Header Copyright Year "
                    f'{copyright_match["modification_year"]} -> '
                    f"{parsed_args.year}"
                )

                return 0
            else:
                print(f"{file}: Licence Header is ok.")
                return 0
    except FileNotFoundError as e:
        print(f"{file}: File is not existing.")
        raise e
    except UnicodeDecodeError as e:
        print(f"{file}: Ignoring binary file.")
        raise e


def _get_exclude_list(
    exclude_file: Path, directories: List[Path]
) -> List[Path]:
    """Tries to get the list of excluded files / directories.
    If a file is given, it will be used. Otherwise it will be searched
    in the executed root path.
    The ignore file should only contain relative paths like *.py,
    not absolute as **/*.py
    """

    if exclude_file is None:
        exclude_file = Path(".pontos-header-ignore")
    try:
        exclude_lines = exclude_file.read_text(encoding="utf-8").split("\n")
    except FileNotFoundError:
        print("No exclude list file found.")
        return []

    expanded_globs = [
        directory.rglob(line.strip())
        for directory in directories
        for line in exclude_lines
        if line
    ]

    exclude_list = []
    for glob_paths in expanded_globs:
        for path in glob_paths:
            if path.is_dir():
                for efile in path.rglob("*"):
                    exclude_list.append(efile.absolute())
            else:
                exclude_list.append(path.absolute())

    return exclude_list


def _parse_args(args=None):
    """Parsing the args"""

    parser = ArgumentParser(
        description="Update copyright in source file headers.",
        prog="pontos-update-header",
    )

    parser.add_argument(
        "--quiet",
        "-q",
        action="store_true",
        help="Don't print messages to the terminal",
    )

    parser.add_argument(
        "--log-file",
        dest="log_file",
        type=str,
        help="Acivate logging using the given file path",
    )

    date_group = parser.add_mutually_exclusive_group()
    date_group.add_argument(
        "-c",
        "--changed",
        action="store_true",
        default=False,
        help=(
            "Update modified year using git log modified year. "
            "This will not changed all files to current year!"
        ),
    )
    date_group.add_argument(
        "-y",
        "--year",
        default=str(datetime.now().year),
        help=(
            "If year is set, modified year will be "
            "set to the specified year."
        ),
    )

    parser.add_argument(
        "-l",
        "--licence",
        choices=SUPPORTED_LICENCES,
        default="GPL-3.0-or-later",
        help=("Use the passed licence type"),
    )

    parser.add_argument(
        "--company",
        default="Greenbone Networks GmbH",
        help=(
            "If a header will be added to file, "
            "it will be licenced by company."
        ),
    )

    files_group = parser.add_mutually_exclusive_group(required=True)
    files_group.add_argument(
        "-f", "--files", nargs="+", help="Files to update."
    )
    files_group.add_argument(
        "-d",
        "--directories",
        nargs="+",
        help="Directories to find files to update recursively.",
    )

    parser.add_argument(
        "--exclude-file",
        help=(
            "File containing glob patterns for files to "
            "ignore when finding files to update in a directory. "
            "Will look for '.pontos-header-ignore' in the directory "
            "if none is given. "
            "The ignore file should only contain relative paths like *.py,"
            "not absolute as **/*.py"
        ),
        type=FileType("r"),
    )

    return parser.parse_args(args)


def main() -> None:
    parsed_args = _parse_args()
    exclude_list = []

    if parsed_args.quiet:
        term = NullTerminal()
    else:
        term = RichTerminal()

    term.bold_info("pontos-update-header")

    if parsed_args.directories:
        if isinstance(parsed_args.directories, list):
            directories = [
                Path(directory) for directory in parsed_args.directories
            ]
        else:
            directories = [Path(parsed_args.directories)]
        # get file paths to exclude
        exclude_list = _get_exclude_list(parsed_args.exclude_file, directories)
        # get files to update
        files = [
            Path(file)
            for directory in directories
            for file in directory.rglob("*")
            if file.is_file()
        ]
    elif parsed_args.files:
        if isinstance(parsed_args.files, list):
            files = [Path(name) for name in parsed_args.files]
        else:
            files = [Path(parsed_args.files)]

    else:
        # should never happen
        term.error("Specify files to update!")
        sys.exit(1)

    regex = re.compile(
        "[Cc]opyright.*?(19[0-9]{2}|20[0-9]{2}) "
        f"?-? ?(19[0-9]{{2}}|20[0-9]{{2}})? ({parsed_args.company})"
    )

    for file in files:
        try:
            if file.absolute() in exclude_list:
                term.warning(f"{file}: Ignoring file from exclusion list.")
            else:
                _update_file(
                    file=file, regex=regex, parsed_args=parsed_args, term=term
                )
        except (FileNotFoundError, UnicodeDecodeError, ValueError):
            continue

    return 0


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pontos',
 'pontos.changelog',
 'pontos.git',
 'pontos.github',
 'pontos.github.actions',
 'pontos.github.api',
 'pontos.github.models',
 'pontos.github.script',
 'pontos.release',
 'pontos.terminal',
 'pontos.testing',
 'pontos.updateheader',
 'pontos.updateheader.templates.AGPL-3.0-or-later',
 'pontos.updateheader.templates.GPL-2.0-or-later',
 'pontos.updateheader.templates.GPL-3.0-or-later',
 'pontos.version',
 'scripts',
 'scripts.github',
 'tests',
 'tests.changelog',
 'tests.git',
 'tests.github',
 'tests.github.actions',
 'tests.github.api',
 'tests.github.models',
 'tests.github.script',
 'tests.release',
 'tests.terminal',
 'tests.testing',
 'tests.updateheader',
 'tests.version']

package_data = \
{'': ['*'], 'pontos.updateheader': ['templates/GPL-2.0-only/*']}

modules = \
['poetry']
install_requires = \
['colorful>=0.5.4,<0.6.0',
 'httpx[http2]>=0.23.0,<0.24.0',
 'packaging>=20.3',
 'rich>=12.4.4,<13.0.0',
 'tomlkit>=0.5.11']

extras_require = \
{':python_version < "3.8"': ['typing-extensions>=4.4.0,<5.0.0']}

entry_points = \
{'console_scripts': ['pontos = pontos:main',
                     'pontos-changelog = pontos.changelog:main',
                     'pontos-github = pontos.github:main',
                     'pontos-github-actions = pontos.github.actions:main',
                     'pontos-github-script = pontos.github.script:main',
                     'pontos-release = pontos.release:main',
                     'pontos-update-header = pontos.updateheader:main',
                     'pontos-version = pontos.version:main']}

setup_kwargs = {
    'name': 'pontos',
    'version': '22.12.0',
    'description': 'Common utilities and tools maintained by Greenbone Networks',
    'long_description': '![Greenbone Logo](https://www.greenbone.net/wp-content/uploads/gb_new-logo_horizontal_rgb_small.png)\n\n# Pontos - Greenbone Python Utilities and Tools <!-- omit in toc -->\n\n[![GitHub releases](https://img.shields.io/github/release/greenbone/pontos.svg)](https://github.com/greenbone/pontos/releases)\n[![PyPI release](https://img.shields.io/pypi/v/pontos.svg)](https://pypi.org/project/pontos/)\n[![code test coverage](https://codecov.io/gh/greenbone/pontos/branch/main/graph/badge.svg)](https://codecov.io/gh/greenbone/pontos)\n[![Build and test](https://github.com/greenbone/pontos/actions/workflows/ci-python.yml/badge.svg)](https://github.com/greenbone/pontos/actions/workflows/ci-python.yml)\n\nThe **pontos** Python package is a collection of utilities, tools, classes and\nfunctions maintained by [Greenbone Networks].\n\nPontos is the German name of the Greek titan [Pontus](https://en.wikipedia.org/wiki/Pontus_(mythology)),\nthe titan of the sea.\n\n## Table of Contents <!-- omit in toc -->\n\n- [Documentation](#documentation)\n- [Installation](#installation)\n  - [Requirements](#requirements)\n  - [Install using pip](#install-using-pip)\n  - [Install using poetry](#install-using-poetry)\n- [Development](#development)\n- [Maintainer](#maintainer)\n- [Contributing](#contributing)\n- [License](#license)\n\n## Documentation\n\nThe documentation for pontos can be found at https://greenbone.github.io/pontos/. Please refer to the documentation for more details as this README just gives a short overview.\n\n## Installation\n\n### Requirements\n\nPython 3.7 and later is supported.\n\n### Install using pip\n\npip 19.0 or later is required.\n\n> **Note**: All commands listed here use the general tool names. If some of\n> these tools are provided by your distribution, you may need to explicitly use\n> the Python 3 version of the tool, e.g. **`pip3`**.\n\nYou can install the latest stable release of **pontos** from the Python\nPackage Index (pypi) using [pip]\n\n    pip install --user pontos\n\n### Install using poetry\n\nBecause **pontos** is a Python library you most likely need a tool to\nhandle Python package dependencies and Python environments. Therefore we\nstrongly recommend using [pipenv] or [poetry].\n\nYou can install the latest stable release of **pontos** and add it as\na dependency for your current project using [poetry]\n\n    poetry add pontos\n\nFor installation via pipenv please take a look at their [documentation][pipenv].\n\n## Development\n\n**pontos** uses [poetry] for its own dependency management and build\nprocess.\n\nFirst install poetry via pip\n\n    pip install --user poetry\n\nAfterwards run\n\n    poetry install\n\nin the checkout directory of **pontos** (the directory containing the\n`pyproject.toml` file) to install all dependencies including the packages only\nrequired for development.\n\nAfterwards activate the git hooks for auto-formatting and linting via\n[autohooks].\n\n    poetry run autohooks activate\n\nValidate the activated git hooks by running\n\n    poetry run autohooks check\n\n## Maintainer\n\nThis project is maintained by [Greenbone Networks GmbH][Greenbone Networks]\n\n## Contributing\n\nYour contributions are highly appreciated. Please\n[create a pull request](https://github.com/greenbone/pontos/pulls)\non GitHub. Bigger changes need to be discussed with the development team via the\n[issues section at GitHub](https://github.com/greenbone/pontos/issues)\nfirst.\n\n## License\n\nCopyright (C) 2020-2022 [Greenbone Networks GmbH][Greenbone Networks]\n\nLicensed under the [GNU General Public License v3.0 or later](LICENSE).\n\n[Greenbone Networks]: https://www.greenbone.net/\n[poetry]: https://python-poetry.org/\n[pip]: https://pip.pypa.io/\n[pipenv]: https://pipenv.pypa.io/\n[autohooks]: https://github.com/greenbone/autohooks\n',
    'author': 'Greenbone Networks GmbH',
    'author_email': 'info@greenbone.net',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/greenbone/pontos/',
    'packages': packages,
    'package_data': package_data,
    'py_modules': modules,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# Copyright (C) 2020-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# pylint: disable=C0413,W0108

import os
import unittest
from contextlib import redirect_stdout
from io import StringIO
from pathlib import Path
from unittest.mock import patch

from pontos import changelog, release
from pontos.git import Git
from pontos.release.helper import get_current_version, version
from pontos.testing import temp_git_repository


class PrepareTestCase(unittest.TestCase):
    def setUp(self) -> None:
        os.environ["GITHUB_TOKEN"] = "foo"
        os.environ["GITHUB_USER"] = "bar"

    @patch("pontos.release.prepare.Git", spec=Git)
    @patch("pontos.release.prepare.Path", spec=Path)
    @patch("pontos.release.helper.version", spec=version)
    @patch("pontos.release.prepare.changelog", spec=changelog)
    def test_prepare_successfully(
        self,
        changelog_mock,
        version_mock,
        _path_mock,
        _git_mock,
    ):
        changelog_mock.update.return_value = ("updated", "changelog")
        version_mock.main.return_value = (True, "MyProject.conf")

        args = [
            "prepare",
            "--project",
            "foo",
            "--release-version",
            "0.0.1",
        ]
        with temp_git_repository(), redirect_stdout(StringIO()):
            released = release.main(
                leave=False,
                args=args,
            )
        self.assertTrue(released)

    @patch(
        "pontos.release.helper.get_current_version", spec=get_current_version
    )
    @patch("pontos.release.prepare.Git", spec=Git)
    @patch("pontos.release.prepare.Path", spec=Path)
    @patch("pontos.release.helper.version", spec=version)
    @patch("pontos.release.prepare.changelog", spec=changelog)
    def test_prepare_calendar_successfully(
        self,
        changelog_mock,
        version_mock,
        _path_mock,
        _git_mock,
        get_current_version_mock,
    ):
        get_current_version_mock.return_value = "21.6.0"
        version_mock.main.return_value = (True, "MyProject.conf")
        changelog_mock.update.return_value = ("updated", "changelog")
        args = [
            "prepare",
            "--project",
            "foo",
            "--calendar",
        ]
        with temp_git_repository(), redirect_stdout(StringIO()):
            released = release.main(
                leave=False,
                args=args,
            )
        self.assertTrue(released)

    @patch("pontos.release.prepare.Git", spec=Git)
    @patch("pontos.release.prepare.Path", spec=Path)
    @patch("pontos.release.helper.version", spec=version)
    @patch("pontos.release.prepare.changelog", spec=changelog)
    def test_use_git_signing_key_on_prepare(
        self,
        changelog_mock,
        version_mock,
        _path_mock,
        git_mock,
    ):
        version_mock.main.return_value = (True, "MyProject.conf")
        changelog_mock.update.return_value = ("updated", "changelog")

        args = [
            "prepare",
            "--project",
            "foo",
            "--git-signing-key",
            "0815",
            "--release-version",
            "0.0.1",
        ]
        with temp_git_repository(), redirect_stdout(StringIO()):
            released = release.main(
                leave=False,
                args=args,
            )

        self.assertTrue(released)
        git_mock.return_value.tag.assert_called_with(
            "v0.0.1", gpg_key_id="0815", message="Automatic release to 0.0.1"
        )
        git_mock.return_value.commit.assert_called_with(
            "Automatic release to 0.0.1", verify=False, gpg_signing_key="0815"
        )

    @patch("pontos.release.prepare.Git", spec=Git)
    @patch("pontos.release.prepare.Path", spec=Path)
    @patch("pontos.release.helper.version", spec=version)
    @patch("pontos.release.prepare.changelog", spec=changelog)
    def test_fail_if_tag_is_already_taken(
        self,
        _changelog_mock,
        version_mock,
        _path_mock,
        git_mock,
    ):
        git_mock.return_value.list_tags.return_value = ["v0.0.1"]

        version_mock.main.return_value = (True, "MyProject.conf")

        args = [
            "prepare",
            "--release-version",
            "0.0.1",
            "--project",
            "bla",
            "--git-signing-key",
            "1337",
        ]

        with temp_git_repository(), self.assertRaises(
            SystemExit
        ), redirect_stdout(StringIO()):
            release.main(
                leave=False,
                args=args,
            )

        git_mock.return_value.list_tags.assert_called_once()

    @patch("pontos.release.prepare.Git", spec=Git)
    @patch("pontos.release.prepare.Path", spec=Path)
    @patch("pontos.release.helper.version", spec=version)
    @patch("pontos.release.prepare.changelog", spec=changelog)
    def test_not_release_when_no_project_found(
        self,
        changelog_mock,
        version_mock,
        _path_mock,
        _git_mock,
    ):
        version_mock.main.return_value = (False, "")
        changelog_mock.update.return_value = ("updated", "changelog")

        args = [
            "prepare",
            "--project",
            "foo",
            "--release-version",
            "0.0.1",
        ]
        with temp_git_repository(), redirect_stdout(StringIO()):
            released = release.main(
                leave=False,
                args=args,
            )
        self.assertFalse(released)

    @patch("pontos.release.prepare.Git", spec=Git)
    @patch("pontos.release.prepare.Path", spec=Path)
    @patch("pontos.release.helper.version", spec=version)
    @patch("pontos.release.prepare.changelog", spec=changelog)
    def test_not_release_when_updating_version_fails(
        self,
        changelog_mock,
        version_mock,
        _path_mock,
        _git_mock,
    ):
        version_mock.main.return_value = (False, "MyProject.conf")
        changelog_mock.update.return_value = ("updated", "changelog")

        args = [
            "prepare",
            "--project",
            "foo",
            "--release-version",
            "0.0.1",
        ]
        with temp_git_repository(), redirect_stdout(StringIO()):
            released = release.main(
                leave=False,
                args=args,
            )
        self.assertFalse(released)

    @patch("pontos.release.prepare.Git", spec=Git)
    @patch("pontos.release.prepare.changelog", spec=changelog)
    @patch("pontos.release.helper.version", spec=version)
    def test_prepare_conventional_commits(
        self,
        version_mock,
        changelog_mock,
        _git_mock,
    ):
        version_mock.main.return_value = (True, "MyProject.conf")

        own_path = Path(__file__).absolute().parent
        with temp_git_repository() as temp_dir, redirect_stdout(StringIO()):
            release_file = temp_dir / ".release.md"

            builder = changelog_mock.ChangelogBuilder.return_value
            builder.create_changelog_file.return_value = own_path / "v1.2.3.md"

            args = [
                "prepare",
                "--project",
                "foo",
                "--release-version",
                "1.2.3",
                "-CC",
            ]
            released = release.main(
                leave=False,
                args=args,
            )

            self.assertTrue(released)

            expected_release_content = """## [21.8.1] - 2021-08-23

## Added

* Need for commits. [1234567](https://github.com/foo/bar/commit/1234567)

## Changed

* fooooo. [1234568](https://github.com/foo/bar/commit/1234568)

[21.8.1]: https://github.com/y0urself/test_workflows/compare/21.8.0...21.8.1"""

            self.assertEqual(
                release_file.read_text(encoding="utf-8"),
                expected_release_content,
            )



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import io
import unittest
from argparse import ArgumentParser
from contextlib import redirect_stderr

from pontos.github.script.errors import GitHubScriptError
from pontos.github.script.load import (
    load_script,
    run_add_arguments_function,
    run_github_script_function,
)
from pontos.testing import temp_file
from tests import IsolatedAsyncioTestCase


class LoadScriptTestCase(unittest.TestCase):
    def test_load_script(self):
        with temp_file(
            "def foo():\n\treturn 1", name="foo.py"
        ) as f, load_script(f) as module:
            self.assertIsNotNone(module)
            self.assertIsNotNone(module.foo)
            self.assertEqual(module.foo(), 1)

    def test_load_script_failure(self):
        with self.assertRaisesRegex(
            ModuleNotFoundError, "No module named 'baz'"
        ):
            with load_script("foo/bar/baz.py"):
                pass


class RunAddArgumentsFunction(unittest.TestCase):
    def test_add_arguments_function(self):
        with temp_file(
            "def add_script_arguments(parser):\n\t"
            "parser.add_argument('--foo')",
            name="foo.py",
        ) as f, load_script(f) as module:
            parser = ArgumentParser()
            run_add_arguments_function(module, parser)
            args = parser.parse_args(["--foo", "123"])
            self.assertEqual(args.foo, "123")

    def test_no_add_arguments_function(self):
        with temp_file(
            "def foo():\n\tpass",
            name="foo.py",
        ) as f, load_script(f) as module:
            parser = ArgumentParser()
            run_add_arguments_function(module, parser)
            with self.assertRaises(SystemExit), redirect_stderr(io.StringIO()):
                parser.parse_args(["--foo", "123"])


class RunGithubScriptFunctionTestCase(unittest.TestCase):
    def test_run_sync_github_script_function(self):
        with temp_file(
            "def github_script(api, args):\n\treturn 1",
            name="foo.py",
        ) as f, load_script(f) as module:
            self.assertEqual(
                run_github_script_function(module, "123", 123, {}), 1
            )

    def test_no_github_script_function(self):
        with temp_file(
            "def foo():\n\tpass",
            name="foo.py",
        ) as f, load_script(f) as module:
            with self.assertRaises(GitHubScriptError):
                run_github_script_function(module, "123", 123, {})


class AsyncRunGithubScriptFunctionTestCase(IsolatedAsyncioTestCase):
    def test_run_async_github_script_function(self):
        with temp_file(
            "async def github_script(api, args):\n\treturn 1",
            name="foo.py",
        ) as f, load_script(f) as module:
            self.assertEqual(
                run_github_script_function(module, "123", 123, {}), 1
            )



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import tempfile
from contextlib import contextmanager
from pathlib import Path
from typing import Generator, Optional

from pontos.git.git import exec_git
from pontos.helper import add_sys_path, ensure_unload_module, unload_module

__all__ = (
    "temp_directory",
    "temp_git_repository",
    "temp_file",
    "temp_python_module",
    "add_sys_path",
    "ensure_unload_module",
    "unload_module",
)


@contextmanager
def temp_directory(
    *, change_into: bool = False, add_to_sys_path: bool = False
) -> Generator[Path, None, None]:
    """
    Context Manager to create a temporary directory

    Args:
        change_into: Set the created temporary as the current working directory.
            The behavior of the current working directory when leaving the
            context manager is undefined.
        add_to_sys_path: Add the created temporary directory to the directories
            for searching for Python modules

    Returns:
        A path to the created temporary directory

    Example:
        .. code-block:: python

            with temp_directory(change_into=True) as tmp:
                new_file = tmp / "test.txt"
    """
    temp_dir = tempfile.TemporaryDirectory()
    dir_path = Path(temp_dir.name)

    if change_into:
        try:
            old_cwd = Path.cwd()
        except FileNotFoundError:
            old_cwd = Path.home()

        os.chdir(dir_path)

    try:
        if add_to_sys_path:
            with add_sys_path(dir_path):
                yield Path(dir_path)
        else:
            yield Path(dir_path)
    finally:
        if change_into:
            try:
                os.chdir(old_cwd)
            finally:
                temp_dir.cleanup()
        else:
            temp_dir.cleanup()


@contextmanager
def temp_git_repository(
    *,
    user_name: str = "Max Mustermann",
    user_email: str = "max.mustermann@example.com",
    branch: str = "main",
) -> Generator[Path, None, None]:
    """
    Context Manager to create a temporary git repository on the filesystem

    Args:
        user_name: User name to configure in the repository.
            Default: Max Mustermann
        user_email: Email address of the user to configure in the repository.
            Default: max.mustermann@example.com
        branch: Branch name to create. Default: main

    Returns:
        A path to the created temporary git repository directory

    Example:
        .. code-block:: python

            with temp_git_repository() as repo:
                new_file = repo / "foo.txt"
                new_file.write_text("Lorem Ipsum")

                exec_git("add", "foo.txt")
    """
    temp_dir = tempfile.TemporaryDirectory()
    temp_path = Path(temp_dir.name)

    try:
        old_cwd = Path.cwd()
    except FileNotFoundError:
        old_cwd = Path.home()

    os.chdir(temp_path)

    exec_git("init", "-b", branch)
    exec_git("config", "--local", "user.email", user_email)
    exec_git("config", "--local", "user.name", user_name)

    try:
        yield temp_path
    finally:
        try:
            os.chdir(old_cwd)
        finally:
            temp_dir.cleanup()


@contextmanager
def temp_file(
    content: str,
    *,
    name: Optional[str] = "test.toml",
    change_into: bool = False,
) -> Generator[Path, None, None]:
    """
    A Context Manager to create a temporary file within a new temporary
    directory. The temporary file and directory are removed when the context is
    exited.

    Args:
        content: Content to write into the temporary file.
        name: Name of the temporary file. "test.toml" by default.
        change_into: Adjust the current working directory to the temporary
            directory.

    Returns:
        A path to the created temporary file

    Example:
        .. code-block:: python

            with temp_file("Lorem Ipsum", name="foo.txt") as fpath:

    """
    with temp_directory(change_into=change_into) as tmp_dir:
        test_file = tmp_dir / name
        test_file.write_text(content, encoding="utf8")
        yield test_file


@contextmanager
def temp_python_module(
    content: str, *, name: str = "foo"
) -> Generator[Path, None, None]:
    """
    A Context Manager to create a new Python module in a temporary directory.
    The temporary directory will be added to the module search path and removed
    from the search path when the context is exited. Also it is ensured that
    the module is unloaded if the context is exited.

    Args:
        content: Python code to write into the temporary module.
        name: Name of the new Python module. By default: "foo".

    Returns:
        A path to the created temporary Python module file

    Example:
        .. code-block:: python

            with temp_python_module(
                "def hello(value):\\n  print(f'Hello {value}')", name="world"
            ) as python_module_path:
                from world import hello
                hello("World")


    """
    with temp_directory(
        add_to_sys_path=True,
    ) as tmp_dir, ensure_unload_module(name):
        test_file = tmp_dir / f"{name}.py"
        test_file.write_text(content, encoding="utf8")
        yield test_file
